@@ -17,11 +17,13 @@
 from __future__ import print_function, unicode_literals
 
 import argparse
+import getpass
 import io
 import json
 import logging
 import signal
 import socket
+import warnings
 
 from os import environ, walk, _exit as os_exit
 from os.path import isfile, isdir, join
@@ -49,7 +51,7 @@ except ImportError:
 
 BASENAME = 'swift'
 commands = ('delete', 'download', 'list', 'post', 'copy', 'stat', 'upload',
-            'capabilities', 'info', 'tempurl', 'auth')
+            'capabilities', 'info', 'tempurl', 'auth', 'bash_completion')
 
 
 def immediate_exit(signum, frame):
@@ -88,7 +90,7 @@ Optional arguments:
 '''.strip("\n")
 
 
-def st_delete(parser, args, output_manager):
+def st_delete(parser, args, output_manager, return_parser=False):
     parser.add_argument(
         '-a', '--all', action='store_true', dest='yes_all',
         default=False, help='Delete all containers and objects.')
@@ -112,6 +114,11 @@ def st_delete(parser, args, output_manager):
         '--container-threads', type=int,
         default=10, help='Number of threads to use for deleting containers. '
         'Its value must be a positive integer. Default is 10.')
+
+    # We return the parser to build up the bash_completion
+    if return_parser:
+        return parser
+
     (options, args) = parse_args(parser, args)
     args = args[1:]
     if (not args and not options['yes_all']) or (args and options['yes_all']):
@@ -279,7 +286,7 @@ Optional arguments:
 '''.strip("\n")
 
 
-def st_download(parser, args, output_manager):
+def st_download(parser, args, output_manager, return_parser=False):
     parser.add_argument(
         '-a', '--all', action='store_true', dest='yes_all',
         default=False, help='Indicates that you really want to download '
@@ -342,6 +349,11 @@ def st_download(parser, args, output_manager):
         'to store the access and modified timestamp for the downloaded file. '
         'With this option, the header is ignored and the timestamps are '
         'created freshly.')
+
+    # We return the parser to build up the bash_completion
+    if return_parser:
+        return parser
+
     (options, args) = parse_args(parser, args)
     args = args[1:]
     if options['out_file'] == '-':
@@ -492,7 +504,7 @@ Optional arguments:
 '''.strip('\n')
 
 
-def st_list(parser, args, output_manager):
+def st_list(parser, args, output_manager, return_parser=False):
 
     def _print_stats(options, stats, human):
         total_count = total_bytes = 0
@@ -569,6 +581,11 @@ def st_list(parser, args, output_manager):
         '-H', '--header', action='append', dest='header',
         default=[],
         help='Adds a custom request header to use for listing.')
+
+    # We return the parser to build up the bash_completion
+    if return_parser:
+        return parser
+
     options, args = parse_args(parser, args)
     args = args[1:]
     if options['delimiter'] and not args:
@@ -627,7 +644,7 @@ Optional arguments:
 '''.strip('\n')
 
 
-def st_stat(parser, args, output_manager):
+def st_stat(parser, args, output_manager, return_parser=False):
     parser.add_argument(
         '--lh', dest='human', action='store_true', default=False,
         help='Report sizes in human readable format similar to ls -lh.')
@@ -636,6 +653,10 @@ def st_stat(parser, args, output_manager):
         default=[],
         help='Adds a custom request header to use for stat.')
 
+    # We return the parser to build up the bash_completion
+    if return_parser:
+        return parser
+
     options, args = parse_args(parser, args)
     args = args[1:]
 
@@ -723,7 +744,7 @@ Optional arguments:
 '''.strip('\n')
 
 
-def st_post(parser, args, output_manager):
+def st_post(parser, args, output_manager, return_parser=False):
     parser.add_argument(
         '-r', '--read-acl', dest='read_acl', help='Read ACL for containers. '
         'Quick summary of ACL syntax: .r:*, .r:-.example.com, '
@@ -748,6 +769,11 @@ def st_post(parser, args, output_manager):
         'This option may be repeated. '
         'Example: -H "content-type:text/plain" '
         '-H "Content-Length: 4000"')
+
+    # We return the parser to build up the bash_completion
+    if return_parser:
+        return parser
+
     (options, args) = parse_args(parser, args)
     args = args[1:]
     if (options['read_acl'] or options['write_acl'] or options['sync_to'] or
@@ -820,7 +846,7 @@ Optional arguments:
 '''.strip('\n')
 
 
-def st_copy(parser, args, output_manager):
+def st_copy(parser, args, output_manager, return_parser=False):
     parser.add_argument(
         '-d', '--destination', help='The container and name of the '
         'destination object')
@@ -837,6 +863,11 @@ def st_copy(parser, args, output_manager):
         'This option may be repeated. '
         'Example: -H "content-type:text/plain" '
         '-H "Content-Length: 4000"')
+
+    # We return the parser to build up the bash_completion
+    if return_parser:
+        return parser
+
     (options, args) = parse_args(parser, args)
     args = args[1:]
 
@@ -891,8 +922,8 @@ def st_copy(parser, args, output_manager):
 st_upload_options = '''[--changed] [--skip-identical] [--segment-size <size>]
                     [--segment-container <container>] [--leave-segments]
                     [--object-threads <thread>] [--segment-threads <threads>]
-                    [--header <header>] [--use-slo] [--ignore-checksum]
-                    [--object-name <object-name>]
+                    [--meta <name:value>] [--header <header>] [--use-slo]
+                    [--ignore-checksum] [--object-name <object-name>]
                     <container> <file_or_directory> [<file_or_directory>] [...]
 '''
 
@@ -928,6 +959,9 @@ Optional arguments:
   --segment-threads <threads>
                         Number of threads to use for uploading object segments.
                         Default is 10.
+  -m, --meta <name:value>
+                        Sets a meta data item. This option may be repeated.
+                        Example: -m Color:Blue -m Size:Large
   -H, --header <header:value>
                         Adds a customized request header. This option may be
                         repeated. Example: -H "content-type:text/plain"
@@ -943,7 +977,9 @@ Optional arguments:
 '''.strip('\n')
 
 
-def st_upload(parser, args, output_manager):
+def st_upload(parser, args, output_manager, return_parser=False):
+    DEFAULT_STDIN_SEGMENT = 10 * 1024 * 1024
+
     parser.add_argument(
         '-c', '--changed', action='store_true', dest='changed',
         default=False, help='Only upload files that have changed since '
@@ -978,6 +1014,10 @@ def st_upload(parser, args, output_manager):
         '--segment-threads', type=int, default=10,
         help='Number of threads to use for uploading object segments. '
         'Its value must be a positive integer. Default is 10.')
+    parser.add_argument(
+        '-m', '--meta', action='append', dest='meta', default=[],
+        help='Sets a meta data item. This option may be repeated. '
+        'Example: -m Color:Blue -m Size:Large')
     parser.add_argument(
         '-H', '--header', action='append', dest='header',
         default=[], help='Set request headers with the syntax header:value. '
@@ -995,6 +1035,11 @@ def st_upload(parser, args, output_manager):
     parser.add_argument(
         '--ignore-checksum', dest='checksum', default=True,
         action='store_false', help='Turn off checksum validation for uploads.')
+
+    # We return the parser to build up the bash_completion
+    if return_parser:
+        return parser
+
     options, args = parse_args(parser, args)
     args = args[1:]
     if len(args) < 2:
@@ -1053,6 +1098,12 @@ def st_upload(parser, args, output_manager):
             st_upload_help)
         return
 
+    if from_stdin:
+        if not options['use_slo']:
+            options['use_slo'] = True
+        if not options['segment_size']:
+            options['segment_size'] = DEFAULT_STDIN_SEGMENT
+
     options['object_uu_threads'] = options['object_threads']
     with SwiftService(options=options) as swift:
         try:
@@ -1168,7 +1219,7 @@ Optional arguments:
 st_info_help = st_capabilities_help
 
 
-def st_capabilities(parser, args, output_manager):
+def st_capabilities(parser, args, output_manager, return_parser=False):
     def _print_compo_cap(name, capabilities):
         for feature, options in sorted(capabilities.items(),
                                        key=lambda x: x[0]):
@@ -1181,6 +1232,11 @@ def st_capabilities(parser, args, output_manager):
 
     parser.add_argument('--json', action='store_true',
                         help='print capability information in json')
+
+    # We return the parser to build up the bash_completion
+    if return_parser:
+        return parser
+
     (options, args) = parse_args(parser, args)
     if args and len(args) > 2:
         output_manager.error('Usage: %s capabilities %s\n%s',
@@ -1229,7 +1285,12 @@ Display auth related authentication variables in shell friendly format.
 '''.strip('\n')
 
 
-def st_auth(parser, args, thread_manager):
+def st_auth(parser, args, thread_manager, return_parser=False):
+
+    # We return the parser to build up the bash_completion
+    if return_parser:
+        return parser
+
     (options, args) = parse_args(parser, args)
     if options['verbose'] > 1:
         if options['auth_version'] in ('1', '1.0'):
@@ -1308,10 +1369,12 @@ Optional arguments:
                         generated.
   --iso8601             If present, the generated temporary URL will contain an
                         ISO 8601 UTC timestamp instead of a Unix timestamp.
+  --ip-range            If present, the temporary URL will be restricted to the
+                        given ip or ip range.
 '''.strip('\n')
 
 
-def st_tempurl(parser, args, thread_manager):
+def st_tempurl(parser, args, thread_manager, return_parser=False):
     parser.add_argument(
         '--absolute', action='store_true',
         dest='absolute_expiry', default=False,
@@ -1331,6 +1394,16 @@ def st_tempurl(parser, args, thread_manager):
         help=("If present, the temporary URL will contain an ISO 8601 UTC "
               "timestamp instead of a Unix timestamp."),
     )
+    parser.add_argument(
+        '--ip-range', action='store',
+        default=None,
+        help=("If present, the temporary URL will be restricted to the "
+              "given ip or ip range."),
+    )
+
+    # We return the parser to build up the bash_completion
+    if return_parser:
+        return parser
 
     (options, args) = parse_args(parser, args)
     args = args[1:]
@@ -1350,7 +1423,8 @@ def st_tempurl(parser, args, thread_manager):
         path = generate_temp_url(parsed.path, timestamp, key, method,
                                  absolute=options['absolute_expiry'],
                                  iso8601=options['iso8601'],
-                                 prefix=options['prefix_based'])
+                                 prefix=options['prefix_based'],
+                                 ip_range=options['ip_range'])
     except ValueError as err:
         thread_manager.error(err)
         return
@@ -1362,6 +1436,66 @@ def st_tempurl(parser, args, thread_manager):
     thread_manager.print_msg(url)
 
 
+st_bash_completion_help = '''Retrieve command specific flags used by bash_completion.
+
+Optional positional arguments:
+  <command>           Swift client command to filter the flags by.
+'''.strip('\n')
+
+
+st_bash_completion_options = '''[command]
+'''
+
+
+def st_bash_completion(parser, args, thread_manager, return_parser=False):
+    if return_parser:
+        return parser
+
+    global commands
+    com = args[1] if len(args) > 1 else None
+
+    if com:
+        if com in commands:
+            fn_commands = ["st_%s" % com]
+        else:
+            print("")
+            return
+    else:
+        fn_commands = [fn for fn in globals().keys()
+                       if fn.startswith('st_') and
+                       not fn.endswith('_options') and
+                       not fn.endswith('_help')]
+
+    subparsers = parser.add_subparsers()
+    subcommands = {}
+    if not com:
+        subcommands['base'] = parser
+    for command in fn_commands:
+        cmd = command[3:]
+        if com:
+            subparser = subparsers.add_parser(
+                cmd, help=globals()['%s_help' % command])
+            add_default_args(subparser)
+            subparser = globals()[command](
+                subparser, args, thread_manager, True)
+            subcommands[cmd] = subparser
+        else:
+            subcommands[cmd] = None
+
+    cmds = set()
+    opts = set()
+    for sc_str, sc in list(subcommands.items()):
+        cmds.add(sc_str)
+        if sc:
+            for option in sc._optionals._option_string_actions:
+                opts.add(option)
+
+    for cmd_to_remove in (com, 'bash_completion', 'base'):
+        if cmd_to_remove in cmds:
+            cmds.remove(cmd_to_remove)
+    print(' '.join(cmds | opts))
+
+
 class HelpFormatter(argparse.HelpFormatter):
     def _format_action_invocation(self, action):
         if not action.option_strings:
@@ -1395,6 +1529,30 @@ class HelpFormatter(argparse.HelpFormatter):
         return action.dest
 
 
+def prompt_for_password():
+    """
+    Prompt the user for a password.
+
+    :raise SystemExit: if a password cannot be entered without it being echoed
+        to the terminal.
+    :return: the entered password.
+    """
+    with warnings.catch_warnings():
+        warnings.filterwarnings('error', category=getpass.GetPassWarning,
+                                append=True)
+        try:
+            # temporarily set signal handling back to default to avoid user
+            # Ctrl-c leaving terminal in weird state
+            signal.signal(signal.SIGINT, signal.SIG_DFL)
+            return getpass.getpass()
+        except EOFError:
+            return None
+        except getpass.GetPassWarning:
+            exit('Input stream incompatible with --prompt option')
+        finally:
+            signal.signal(signal.SIGINT, immediate_exit)
+
+
 def parse_args(parser, args, enforce_requires=True):
     options, args = parser.parse_known_args(args or ['-h'])
     options = vars(options)
@@ -1420,6 +1578,10 @@ def parse_args(parser, args, enforce_requires=True):
     if args and args[0] == 'tempurl':
         return options, args
 
+    # do this before process_options sets default auth version
+    if enforce_requires and options['prompt']:
+        options['key'] = options['os_password'] = prompt_for_password()
+
     # Massage auth version; build out os_options subdict
     process_options(options)
 
@@ -1454,88 +1616,7 @@ adding "-V 2" is necessary for this.'''.strip('\n'))
     return options, args
 
 
-def main(arguments=None):
-    argv = sys_argv if arguments is None else arguments
-
-    argv = [a if isinstance(a, text_type) else a.decode('utf-8') for a in argv]
-
-    version = client_version
-    parser = argparse.ArgumentParser(
-        add_help=False, formatter_class=HelpFormatter, usage='''
-%(prog)s [--version] [--help] [--os-help] [--snet] [--verbose]
-             [--debug] [--info] [--quiet] [--auth <auth_url>]
-             [--auth-version <auth_version> |
-                 --os-identity-api-version <auth_version> ]
-             [--user <username>]
-             [--key <api_key>] [--retries <num_retries>]
-             [--os-username <auth-user-name>] [--os-password <auth-password>]
-             [--os-user-id <auth-user-id>]
-             [--os-user-domain-id <auth-user-domain-id>]
-             [--os-user-domain-name <auth-user-domain-name>]
-             [--os-tenant-id <auth-tenant-id>]
-             [--os-tenant-name <auth-tenant-name>]
-             [--os-project-id <auth-project-id>]
-             [--os-project-name <auth-project-name>]
-             [--os-project-domain-id <auth-project-domain-id>]
-             [--os-project-domain-name <auth-project-domain-name>]
-             [--os-auth-url <auth-url>] [--os-auth-token <auth-token>]
-             [--os-storage-url <storage-url>] [--os-region-name <region-name>]
-             [--os-service-type <service-type>]
-             [--os-endpoint-type <endpoint-type>]
-             [--os-cacert <ca-certificate>] [--insecure]
-             [--os-cert <client-certificate-file>]
-             [--os-key <client-certificate-key-file>]
-             [--no-ssl-compression]
-             <subcommand> [--help] [<subcommand options>]
-
-Command-line interface to the OpenStack Swift API.
-
-Positional arguments:
-  <subcommand>
-    delete               Delete a container or objects within a container.
-    download             Download objects from containers.
-    list                 Lists the containers for the account or the objects
-                         for a container.
-    post                 Updates meta information for the account, container,
-                         or object; creates containers if not present.
-    copy                 Copies object, optionally adds meta
-    stat                 Displays information for the account, container,
-                         or object.
-    upload               Uploads files or directories to the given container.
-    capabilities         List cluster capabilities.
-    tempurl              Create a temporary URL.
-    auth                 Display auth related environment variables.
-
-Examples:
-  %(prog)s download --help
-
-  %(prog)s -A https://api.example.com/v1.0 \\
-      -U user -K api_key stat -v
-
-  %(prog)s --os-auth-url https://api.example.com/v2.0 \\
-      --os-tenant-name tenant \\
-      --os-username user --os-password password list
-
-  %(prog)s --os-auth-url https://api.example.com/v3 --auth-version 3\\
-      --os-project-name project1 --os-project-domain-name domain1 \\
-      --os-username user --os-user-domain-name domain1 \\
-      --os-password password list
-
-  %(prog)s --os-auth-url https://api.example.com/v3 --auth-version 3\\
-      --os-project-id 0123456789abcdef0123456789abcdef \\
-      --os-user-id abcdef0123456789abcdef0123456789 \\
-      --os-password password list
-
-  %(prog)s --os-auth-token 6ee5eb33efad4e45ab46806eac010566 \\
-      --os-storage-url https://10.1.5.2:8080/v1/AUTH_ced809b6a4baea7aeab61a \\
-      list
-
-  %(prog)s list --lh
-'''.strip('\n'))
-    parser.add_argument('--version', action='version',
-                        version='python-swiftclient %s' % version)
-    parser.add_argument('-h', '--help', action='store_true')
-
+def add_default_args(parser):
     default_auth_version = '1.0'
     for k in ('ST_AUTH_VERSION', 'OS_AUTH_VERSION', 'OS_IDENTITY_API_VERSION'):
         try:
@@ -1595,6 +1676,17 @@ Examples:
                         help='This option is deprecated and not used anymore. '
                              'SSL compression should be disabled by default '
                              'by the system SSL library.')
+    parser.add_argument('--force-auth-retry',
+                        action='store_true', dest='force_auth_retry',
+                        default=False,
+                        help='Force a re-auth attempt on '
+                             'any error other than 401 unauthorized')
+    parser.add_argument('--prompt',
+                        action='store_true', dest='prompt',
+                        default=False,
+                        help='Prompt user to enter a password which overrides '
+                             'any password supplied via --key, --os-password '
+                             'or environment variables.')
 
     os_grp = parser.add_argument_group("OpenStack authentication options")
     os_grp.add_argument('--os-username',
@@ -1737,6 +1829,100 @@ Examples:
                         default=environ.get('OS_KEY'),
                         help='Specify a client certificate key file (for '
                         'client auth). Defaults to env[OS_KEY].')
+
+
+def main(arguments=None):
+    argv = sys_argv if arguments is None else arguments
+
+    argv = [a if isinstance(a, text_type) else a.decode('utf-8') for a in argv]
+
+    parser = argparse.ArgumentParser(
+        add_help=False, formatter_class=HelpFormatter, usage='''
+%(prog)s [--version] [--help] [--os-help] [--snet] [--verbose]
+             [--debug] [--info] [--quiet] [--auth <auth_url>]
+             [--auth-version <auth_version> |
+                 --os-identity-api-version <auth_version> ]
+             [--user <username>]
+             [--key <api_key>] [--retries <num_retries>]
+             [--os-username <auth-user-name>]
+             [--os-password <auth-password>]
+             [--os-user-id <auth-user-id>]
+             [--os-user-domain-id <auth-user-domain-id>]
+             [--os-user-domain-name <auth-user-domain-name>]
+             [--os-tenant-id <auth-tenant-id>]
+             [--os-tenant-name <auth-tenant-name>]
+             [--os-project-id <auth-project-id>]
+             [--os-project-name <auth-project-name>]
+             [--os-project-domain-id <auth-project-domain-id>]
+             [--os-project-domain-name <auth-project-domain-name>]
+             [--os-auth-url <auth-url>]
+             [--os-auth-token <auth-token>]
+             [--os-storage-url <storage-url>]
+             [--os-region-name <region-name>]
+             [--os-service-type <service-type>]
+             [--os-endpoint-type <endpoint-type>]
+             [--os-cacert <ca-certificate>]
+             [--insecure]
+             [--os-cert <client-certificate-file>]
+             [--os-key <client-certificate-key-file>]
+             [--no-ssl-compression]
+             [--force-auth-retry]
+             <subcommand> [--help] [<subcommand options>]
+
+Command-line interface to the OpenStack Swift API.
+
+Positional arguments:
+  <subcommand>
+    delete               Delete a container or objects within a container.
+    download             Download objects from containers.
+    list                 Lists the containers for the account or the objects
+                         for a container.
+    post                 Updates meta information for the account, container,
+                         or object; creates containers if not present.
+    copy                 Copies object, optionally adds meta
+    stat                 Displays information for the account, container,
+                         or object.
+    upload               Uploads files or directories to the given container.
+    capabilities         List cluster capabilities.
+    tempurl              Create a temporary URL.
+    auth                 Display auth related environment variables.
+    bash_completion      Outputs option and flag cli data ready for
+                         bash_completion.
+
+Examples:
+  %(prog)s download --help
+
+  %(prog)s -A https://api.example.com/v1.0 \\
+      -U user -K api_key stat -v
+
+  %(prog)s --os-auth-url https://api.example.com/v2.0 \\
+      --os-tenant-name tenant \\
+      --os-username user --os-password password list
+
+  %(prog)s --os-auth-url https://api.example.com/v3 --auth-version 3\\
+      --os-project-name project1 --os-project-domain-name domain1 \\
+      --os-username user --os-user-domain-name domain1 \\
+      --os-password password list
+
+  %(prog)s --os-auth-url https://api.example.com/v3 --auth-version 3\\
+      --os-project-id 0123456789abcdef0123456789abcdef \\
+      --os-user-id abcdef0123456789abcdef0123456789 \\
+      --os-password password list
+
+  %(prog)s --os-auth-token 6ee5eb33efad4e45ab46806eac010566 \\
+      --os-storage-url https://10.1.5.2:8080/v1/AUTH_ced809b6a4baea7aeab61a \\
+      list
+
+  %(prog)s list --lh
+'''.strip('\n'))
+
+    version = client_version
+    parser.add_argument('--version', action='version',
+                        version='python-swiftclient %s' % version)
+    parser.add_argument('-h', '--help', action='store_true')
+
+    add_default_args(parser)
+
     options, args = parse_args(parser, argv[1:], enforce_requires=False)
 
     if options['help'] or options['os_help']:
@@ -1757,9 +1943,14 @@ Examples:
         parser.usage = globals()['st_%s_help' % args[0]]
         if options['insecure']:
             import requests
-            from requests.packages.urllib3.exceptions import \
-                InsecureRequestWarning
-            requests.packages.urllib3.disable_warnings(InsecureRequestWarning)
+            try:
+                from requests.packages.urllib3.exceptions import \
+                    InsecureRequestWarning
+            except ImportError:
+                pass
+            else:
+                requests.packages.urllib3.disable_warnings(
+                    InsecureRequestWarning)
         try:
             globals()['st_%s' % args[0]](parser, argv[1:], output)
         except ClientException as err:
